from time import time

from funki import λ
from funki.terminal import Terminal
from funki import report
from funki import text_colours as tc

default_equality_operator = lambda y, z: y==z

class Funk:
  def _default_run(self, σ, x):
    print()
    raise NotImplementedError(
      f"please specify run for funk: '{tc.info(self.name)}'"
    )

  def __init__(
    self,
    name:str,
    io_value_pairs=[],
    output_properties=[],
    description='',
    setup=λ.default_setup,
    run=None,
    tear_down=λ.nop_x,
    custom_tests=None,
    special_return_equality_operator=None
  ):
    self.name = name
    self._state = {'term': Terminal()}
    self.description = description
    self._run = self._default_run if run is None else run
    self._tear_down = tear_down

    setup_result = setup()
    if setup_result is None:
      setup_result = {}

    for key in setup_result.keys():
      self._state[key] = setup_result[key]

    if all([
      not io_value_pairs,
      not output_properties,
      not custom_tests
    ]):
      raise ValueError(
        'Must specify io_value pairs, output properties, or custom_tests.'
      )

    self.io_value_pairs = io_value_pairs
    self._output_properties = output_properties
    self.custom_tests = custom_tests

    if special_return_equality_operator is not None:
      self.equality_operator = special_return_equality_operator
    else:
      self.equality_operator = default_equality_operator
    
  run = lambda self, x=None: self._run(self._state, x)

  def _test(self):
    """
    Input x provided to function f returns obervation z
    x --> f(x) --> z

    Input x provided to function f is expected to return y
    x --> f(x) --> y

    Test passes is expected return value y is equal to observation z.

    x includes:
      x_args: arguments passed to the function as single value or dict of values
      x_stdi: stdi stream provided to the function while running
    
    y includes:
      y_return: the expected return value from calling f(x)
      y_stdo: the expected stdo data emitted by f(x) during it's execution.
    
    z includes:
      z_return: the observed return value from calling f(x)
      z_stdo: the observed stdo data emitted by f(x) during it's execution.
    """

    test_start_time = time()

    # For each z:

    io_pairs = (
      [({}, {})] if len(self.io_value_pairs) <= 0 else self.io_value_pairs
    )

    for io_pair in io_pairs:
      self._state['term'].reset_stream_lists()

      i, o = io_pair

      x_args = None if 'x' not in i else i['x']
      x_stdi = [] if 'stdi' not in i else i['stdi']
      self._state['term'].input_stream_as_list += x_stdi # Load into test terminal
      
      y_return = None if 'y' not in o else o['y']
      y_stdo = [] if 'stdo' not in o else o['stdo']

      z_return = self.run(x_args)
      z_stdo = self._state['term'].output_stream_as_list

      xyz = {
        'x_args': x_args,
        'x_stdi': x_stdi,
        'y_return': y_return,
        'y_stdo': y_stdo,
        'z_return': z_return,
        'z_stdo': z_stdo,
      }

      # Only do this comparison if the expected return is not none.
      if y_return is not None:
        # Test y_return == z_return
        if not self.equality_operator(y_return, z_return):
        # if y_return != z_return:
          return report.fail(
            self.name,
            xyz,
            'y_return != z_return',
            test_start_time
          )

      # Test y_stdo == z_stdo
      if y_stdo != z_stdo:
        return report.fail(self.name, xyz, 'y_stdo != z_stdo', test_start_time)

      # Test z_return satisfies output properties
      for output_property in self._output_properties:
        if isinstance(output_property, Funk):
          if not output_property.run(z_return):
            return report.property_failure(
              self.name,
              f'{output_property.name}({z_return})'
            )
        else:
          raise TypeError('property must be written as a Funk')

    if self.custom_tests is not None:
      for (custom_test_index, custom_test) in enumerate(self.custom_tests):
        if not custom_test():
          return report.fail(
            self.name,
            xyz,
            mode=f'custom_tests[{custom_test_index}]()',
            time_started=test_start_time
          )

    return report.success(self.name, test_start_time)

  def test(self):
    self._state['term'].mode = 'test'

    # Start Timer
    print(f'{tc.warning("TESTING")} {tc.info(self.name)} ', end='')
    result = self._test()


    self._state['term'].mode = 'run'
    return result

  __del__ = lambda self: self._tear_down(self._state)

  __call__ = lambda self, x=None: self._run(self._state, x)

  # ###########################################################################

  # if len(self.io_value_pairs) <= 0:
  #   # Check that the output satisfies the output property constraints if
  #   # no input is supplied.
  #   if self._output_properties:
  #     z_return = self.run({})

  #   for output_property in self._output_properties:
  #     has_output_property = output_property(z_return)
  #     if not has_output_property:
  #       return report.property_failure(self.name, z_return)

  # for _io_pair in self.io_value_pairs:
  #   self._state['term'].reset_stdo()

  #   __input, _ = _io_pair

  #   if 'stdi' not in __input.keys():
  #     __input['stdi'] = []

  #   if isinstance(__input['stdi'], list):
  #     self._state['term'].input_stream_as_list += __input['stdi']
    
  #   xyz = self._load_xyz(_io_pair)

  #   if self.observation_does_not_match_expectation(xyz):
  #     return report.fail(self.name, xyz)

  #   for output_property in self._output_properties:
  #     z_return = xyz['z_return']
  #     has_output_property = output_property(z_return)
  #     if not has_output_property:
  #       return report.property_failure(self.name, z_return)
  # return report.success(self.name)

  # def observation_does_not_match_expectation(self, xyz):
  #   if any([
  #     isinstance(xyz['y_return'], DataFrame),
  #     isinstance(xyz['z_return'], DataFrame)
  #   ]):
  #     # Dealing with a DataFrame
  #     if not xyz['y_return'].equals(xyz['z_return']):
  #       print(xyz['y_return'].compare(xyz['z_return']))

  #     return any([
  #       not xyz['y_return'].equals(xyz['z_return']),
  #       xyz['z_stdo'] != xyz['y_stdo']
  #     ])
  #   else:
  #     # Not dealing with a DataFrame
  #     if xyz['y_return'] == "DO_NOT_CARE":
  #       return False # Ignore this case as it is marked as DO_NOT_CARE
  #     else:
  #       return any([
  #         xyz['z_return'] != xyz['y_return'],
  #         xyz['z_stdo'] != xyz['y_stdo']
  #       ])

  # def _load_xyz(self, io_pair):
  #     __input, __target = io_pair
  #     xyz ={}
      
  #     if 'x' not in __input.keys():
  #       __input['x'] = None
  #     xyz['x_args']   = __input['x']

  #     if 'stdo' not in __target.keys():
  #       __target['stdo'] = []
  #     xyz['y_stdo'] = __target['stdo']
      
  #     if 'y' not in __target.keys():
  #       __target['y'] = "DO_NOT_CARE"
  #     xyz['y_return'] = __target['y']

  #     xyz['z_return'] = self.run(xyz['x_args'])
  #     xyz['z_stdo'] = self._state['term'].output_stream_as_list
  #     xyz['x_stdi']  = __input['stdi']

  #     if 'file' in __target.keys():
  #       xyz['y_file'] = __target['file']['content']
  #       if 'read_as' in __target['file']:
  #         if __target['file']['read_as'] == 'xlsx as df':
  #           xyz['z_file'] = read_excel(__target['file']['filename'])
  #         elif __target['file']['read_as'] == 'pickle':
  #           with open(__target['file']['filename'], 'rb') as φ:
  #             xyz['z_file'] = pickle.load(φ)
  #         else:
  #           raise NotImplementedError('Need to write this branch')
  #       else:
  #         raise NotImplementedError('Need to write this branch')
  #     return xyz
